/* rwsem-spinlock.c: R/W semaphores: contention handling functions for generic spinlock
 *                                   implementation
 *
 * Copyright (c) 2001   David Howells (dhowells@redhat.com).
 */
#include <linux/rwsem.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/compiler.h>

struct rwsem_waiter {
	struct rwsem_waiter	*next;
	struct task_struct	*task;
	unsigned int		flags;
#define RWSEM_WAITING_FOR_READ	0x00000001
#define RWSEM_WAITING_FOR_WRITE	0x00000002
};

#if RWSEM_DEBUG
void rwsemtrace(struct rw_semaphore *sem, const char *str)
{
	if (sem->debug)
		printk("[%d] %s({%d,%d})\n",current->pid,str,sem->active,sem->waiting);
}
#endif

/*
 * initialise the semaphore
 */
void init_rwsem(struct rw_semaphore *sem)
{
	sem->active = 0;
	sem->waiting = 0;
	spin_lock_init(&sem->wait_lock);
	sem->wait_front = NULL;
	sem->wait_back = &sem->wait_front;
#if RWSEM_DEBUG
	sem->debug = 0;
#endif
}

/*
 * handle the lock being released whilst there are processes blocked on it that can now run
 * - if we come here, then:
 *   - the 'active count' _reached_ zero
 *   - the 'waiting count' is non-zero
 * - the spinlock must be held by the caller
 * - woken process blocks are discarded from the list after having flags zeroised
 */
static inline struct rw_semaphore *__rwsem_do_wake(struct rw_semaphore *sem)
{
	struct rwsem_waiter *waiter, *next;
	int woken, loop;

	rwsemtrace(sem,"Entering __rwsem_do_wake");

	waiter = sem->wait_front;

	if (__builtin_expect(!waiter,0))
	  goto list_unexpectedly_empty;

	next = NULL;

	/* try to grant a single write lock if there's a writer at the front of the queue
	 * - we leave the 'waiting count' incremented to signify potential contention
	 */
	if (waiter->flags & RWSEM_WAITING_FOR_WRITE) {
		sem->active++;
		next = waiter->next;
		waiter->flags = 0;
		wake_up_process(waiter->task);
		goto discard_woken_processes;
	}

	/* grant an infinite number of read locks to the readers at the front of the queue */
	woken = 0;
	do {
		woken++;
		waiter = waiter->next;
	} while (waiter && waiter->flags&RWSEM_WAITING_FOR_READ);

	sem->active += woken;
	sem->waiting -= woken;

	waiter = sem->wait_front;
	for (loop=woken; loop>0; loop--) {
		next = waiter->next;
		waiter->flags = 0;
		wake_up_process(waiter->task);
		waiter = next;
	}

 discard_woken_processes:
	sem->wait_front = next;
	if (!next) sem->wait_back = &sem->wait_front;

 out:
	rwsemtrace(sem,"Leaving __rwsem_do_wake");
	return sem;

 list_unexpectedly_empty:
	printk("__rwsem_do_wake(): wait_list unexpectedly empty\n");
	printk("[%d] %p = { %d, %d })\n",current->pid,sem,sem->active,sem->waiting);
	BUG();
	goto out;
}

/*
 * get a read lock on the semaphore
 */
void __down_read(struct rw_semaphore *sem)
{
	struct rwsem_waiter waiter;
	struct task_struct *tsk = current;

	rwsemtrace(sem,"Entering __down_read");

	spin_lock(&sem->wait_lock);

	if (!sem->waiting) {
		/* granted */
		sem->active++;
		spin_unlock(&sem->wait_lock);
		goto out;
	}
	sem->waiting++;

	set_task_state(tsk,TASK_UNINTERRUPTIBLE);

	/* set up my own style of waitqueue */
	waiter.next = NULL;
	waiter.task = tsk;
	waiter.flags = RWSEM_WAITING_FOR_READ;

	*sem->wait_back = &waiter; /* add to back of queue */
	sem->wait_back = &waiter.next;

	/* we don't need to touch the semaphore struct anymore */
	spin_unlock(&sem->wait_lock);

	/* wait to be given the lock */
	for (;;) {
		if (!waiter.flags)
			break;
		schedule();
		set_task_state(tsk, TASK_UNINTERRUPTIBLE);
	}

	tsk->state = TASK_RUNNING;

 out:
	rwsemtrace(sem,"Leaving __down_read");
}

/*
 * get a write lock on the semaphore
 * - note that we increment the waiting count anyway to indicate an exclusive lock
 */
void __down_write(struct rw_semaphore *sem)
{
	struct rwsem_waiter waiter;
	struct task_struct *tsk = current;

	rwsemtrace(sem,"Entering __down_write");

	spin_lock(&sem->wait_lock);

	if (!sem->waiting && !sem->active) {
		/* granted */
		sem->active++;
		sem->waiting++;
		spin_unlock(&sem->wait_lock);
		goto out;
	}
	sem->waiting++;

	set_task_state(tsk,TASK_UNINTERRUPTIBLE);

	/* set up my own style of waitqueue */
	waiter.next = NULL;
	waiter.task = tsk;
	waiter.flags = RWSEM_WAITING_FOR_WRITE;

	*sem->wait_back = &waiter; /* add to back of queue */
	sem->wait_back = &waiter.next;

	/* we don't need to touch the semaphore struct anymore */
	spin_unlock(&sem->wait_lock);

	/* wait to be given the lock */
	for (;;) {
		if (!waiter.flags)
			break;
		schedule();
		set_task_state(tsk, TASK_UNINTERRUPTIBLE);
	}

	tsk->state = TASK_RUNNING;

 out:
	rwsemtrace(sem,"Leaving __down_write");
}

/*
 * release a read lock on the semaphore
 */
void __up_read(struct rw_semaphore *sem)
{
	rwsemtrace(sem,"Entering __up_read");

	spin_lock(&sem->wait_lock);

	if (--sem->active==0 && sem->waiting)
		__rwsem_do_wake(sem);

	spin_unlock(&sem->wait_lock);

	rwsemtrace(sem,"Leaving __up_read");
}

/*
 * release a write lock on the semaphore
 */
void __up_write(struct rw_semaphore *sem)
{
	rwsemtrace(sem,"Entering __up_write");

	spin_lock(&sem->wait_lock);

	sem->waiting--;
	if (--sem->active==0 && sem->waiting)
		__rwsem_do_wake(sem);

	spin_unlock(&sem->wait_lock);

	rwsemtrace(sem,"Leaving __up_write");
}

EXPORT_SYMBOL(init_rwsem);
EXPORT_SYMBOL(__down_read);
EXPORT_SYMBOL(__down_write);
EXPORT_SYMBOL(__up_read);
EXPORT_SYMBOL(__up_write);
#if RWSEM_DEBUG
EXPORT_SYMBOL(rwsemtrace);
#endif
